# FDR Main Client Module

# Libraries
from .utils.utils import *
from .utils.login import *
from .utils.cmd_utils.cmd_main import *
from pwinput import pwinput
import multiprocessing

# FDR Client Method
def fdr_client():

    # Welcome message
    welcome_message()

    # Login
    un = input("Username: ")
    pw = pwinput("Password: ")
    
    if auth(un, pw) == 200:

        # cmd options after login
        initial_command_options()

    else:
        auth_denied()

# Main Executable
if __name__ == "__main__":

    # import multiprocessing freeze
    multiprocessing.freeze_support()

    # run client
    fdr_client()


from datetime import datetime

class CommandLogs:
    def __init__(self):
        self.history = []

    def add_text(self, text):
        self.history.append((datetime.now(), text))

    def get_history(self):
        formatted_history = []
        for timestamp, text in self.history:
            formatted_time = timestamp.strftime("%Y-%m-%d %H:%M:%S")
            formatted_history.append(f"[{formatted_time}] {text}")
        return "\n".join(formatted_history)


def main():
    history = CommandLogs()
    
    while True:
        user_input = input("Enter text: ")
        
        if user_input == "fdr log":
            print("Text History:")
            print(history.get_history())
        else:
            history.add_text(user_input[3:])


if __name__ == "__main__":
    main()



# Libraries
from tqdm import tqdm
from colorama import init, Fore

# Initialize colorama
init(autoreset=True)

# Custom Progress Bar Class
class CustomTqdm(tqdm):
    def update(self, n=1):
        super().update(n)
        if self.n == self.total:
            self.set_description(Fore.GREEN + "DONE" + Fore.RESET, refresh=True) 
            self.set_postfix() 
        else:
            self.set_description(Fore.RED + "LOADING" + Fore.RESET, refresh=True) 
            self.set_postfix()  


# Libraries
from ...utils.utils import print_lines_per_terminal_size
from ...utils.cmd_utils.cmd_main import clear
from ...classes.DatasetManager import DatasetManager
from termcolor import cprint
from ...utils.utils import dir_picking, check_dvc_file, sample_headers, sample_data
from ...utils.dvc_utils.dvc_ops import dvc_pipeline

# FDR Information Section
def get_operations_info():
    """
    Display download actions info section
    """
    print_lines_per_terminal_size()
    print("Useful commands:")
    print_lines_per_terminal_size()
    print("- fdr pull   -->  download dataset    ,args: [--name/-n]")
    print("- fdr push   -->  upload dataset      ,args: None")
    print("- fdr show   -->  show all datasets   ,args: None")
    print("- fdr doc    -->  show documentation  ,args: None")
    print("- fdr clear  -->  clear fdr terminal  ,args: None")
    print("- fdr menu   -->  exit to menu        ,args: None")
    print("- fdr log    -->  show history        ,args: [--save/-s]")
    print_lines_per_terminal_size()

def fdr_clear(page_title:str):
    """
    FDR clear command
    """
    clear()
    cprint(f'\n{page_title}\n', 'cyan', attrs=['bold'])
    get_operations_info()

def fdr_log_save():
    """
    FDR log command --save||-s
    """
    # dir picking:
    folder_path = dir_picking('Please select a dataset folder to save the history of FDR commands')
    
    # return non-null dir
    if not folder_path:
        print("No folder selected. Operation canceled.")
        return None
    else:
        print(f"Saved successfully at: '{folder_path}/fdr_logs.txt'")

    return folder_path

def fdr_push():
    """
    FDR push command
    """
    # Pick dataset dir
    folder_path = dir_picking(message='Please select a dataset folder to upload to remote FDR repository')

    # DVC push
    if folder_path:
        cprint(f'\nPushing "{folder_path}" to FaradAI remote storage', "yellow", attrs=['bold'])
        dvc_pipeline(folder_path, action="push", dvc_image=None)
        cprint(f'Done!', "green", attrs=['bold'])
    else:
        cprint("No folder selected. Operation canceled.", "red", attrs=['bold'])

def fdr_pull(args):
    """
    FDR pull command
    """
    # init name
    dataset_name = None

    # handle dir picking and dvc
    if args:
        if '--name' in args or '-n' in args:
            name_index = args.index('--name') if '--name' in args else args.index('-n')
            if len(args) >= name_index + 2:
                dataset_name = args[name_index + 1]

                # Check dataset availability
                manager = DatasetManager(data=sample_data(), headers=sample_headers())
                dataset_names = manager.get_dataset_names()

                if dataset_name in dataset_names:

                    # Choose dir to save dataset
                    folder_path = dir_picking(message='Select a target folder to download dataset from remote FDR repository')

                    if folder_path and check_dvc_file(folder_path):
                    # DVC pull --name
                        cprint(f'\nPulling "{dataset_name}" from FaradAI remote storage to "{folder_path}"', "yellow", attrs=['bold'])
                        dvc_pipeline(folder_path, action="pull", dvc_image=dataset_name)
                        cprint(f'Done!', "green", attrs=['bold'])
                    elif not folder_path:
                        cprint("No folder selected. Operation canceled.", "red", attrs=['bold'])
                    elif not check_dvc_file(folder_path):
                        cprint(".dvc file does not exist or is corrupt.", "red", attrs=['bold'])

                else:
                    cprint(f'Dataset does not exist in remote FDR repository or access is denied', 
                           "red", 
                           attrs=['bold'])

    # return dataset name
    if dataset_name:
        return dataset_name

    else:
        print("Invalid argument. Please specify dataset name [--name/-n]")
        return None


import os
import time
import subprocess
from tqdm import tqdm
from ...utils.utils import progress_update

def dvc_init(parent_dir):
    """
    Initialize DVC in current wd
    """
    dvc_init_cmd = ['dvc', 'init', '--no-scm', '-f']
    subprocess.run(dvc_init_cmd, check=True, cwd=parent_dir, shell=True, 
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)

def dvc_analytics_off(parent_dir):
    """
    Disable DVC analytics
    """
    dvc_analytics_off_cmd = ['dvc', 'config', 'core.analytics', 'false']
    subprocess.run(dvc_analytics_off_cmd, check=True, cwd=parent_dir, 
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)


def dvc_remote(parent_dir):
    """
    Set DVC remote
    """
    dvc_add_remote = ["dvc", "remote", "add", "--default", "drive",
                      "gdrive://1YVePRVMrdkyKQgzBci_vLNhhw6GnAE8M"]
    subprocess.run(dvc_add_remote, check=True, cwd=parent_dir, 
                       stdout=subprocess.DEVNULL)
    
    dvc_acknowledge_abuse = ["dvc", "remote", "modify", "drive",
                             "gdrive_acknowledge_abuse", "true"]
    subprocess.run(dvc_acknowledge_abuse, check=True, cwd=parent_dir, 
                       stdout=subprocess.DEVNULL)

def dvc_add(parent_dir, dir):
    """
    Add target to DVC cache
    """
    # Add
    dvc_add = ["dvc", "add", f"{dir}"]
    subprocess.run(dvc_add, check=True, cwd=parent_dir, shell=True,
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)
    
def dvc_push(parent_dir, workers):
    """
    Push to DVC remote
    """
    # Push
    dvc_push = ["dvc", "push", f"-j {workers}"]
    subprocess.run(dvc_push, check=True, cwd=parent_dir, shell=True,
                        stdout=subprocess.DEVNULL, 
                        stderr=subprocess.DEVNULL)
    
def dvc_pull(dir, dvc_image):
    """
    Pull dataset from remote based on .dvc file
    """
    dvc_fetch = ["dvc", "fetch", f'{dvc_image}.dvc']
    subprocess.run(dvc_fetch, check=True, cwd=dir, shell=True,
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)
    dvc_pull = ["dvc", "checkout", f'{dvc_image}.dvc']
    subprocess.run(dvc_pull, check=True, cwd=dir, shell=True,
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)

def dvc_destroy(parent_dir):
    """
    Uninitialize DVC in current wd by force
    """
    dvc_destroy = ["dvc", "destroy", "--force"]
    subprocess.run(dvc_destroy, check=True, cwd=parent_dir, shell=True,
                       stdout=subprocess.DEVNULL, 
                       stderr=subprocess.DEVNULL)
    
def dvc_retry_with_backoff(func, max_retries=5, retry_delay=10):
    """
    Retry mechanism for dvc methods
    """
    for retry in range(max_retries):
        try:
            func()
            break
        except Exception as e:
            print(f"Attempt {retry + 1} failed. Error: {e}")
            if retry < max_retries - 1:
                print(f"Retrying in {retry_delay} seconds...")
                time.sleep(retry_delay)
            else:
                print(f"Maximum retries ({max_retries}) reached.")
                raise

def dvc_pipeline(target_dir, action, dvc_image):
    """
    Main DVC Operation Pipeline
    """
    with tqdm(total=100, unit="%") as pbar:

        # initial pbar update
        progress_update(pbar, 5)

        # get cwd & parent
        parent_dir = os.path.dirname(target_dir)
        progress_update(pbar, 5)

        # dvc init
        dvc_init(parent_dir)
        progress_update(pbar, 5)

        # dvc disable analytics
        dvc_analytics_off(parent_dir)
        progress_update(pbar, 5)
        
        # dvc remote
        dvc_remote(parent_dir)
        progress_update(pbar, 5)

        # dvc action
        if action == "push":
            # add
            dvc_add(parent_dir, target_dir)
            # push & retry mechanism
            dvc_retry_with_backoff(lambda: dvc_push(parent_dir,
                                                    workers=8))

            
        elif action == "pull":
            # pull & retry mechanism
            dvc_retry_with_backoff(lambda: dvc_pull(target_dir, 
                                                    dvc_image))

        # dvc destroy
        progress_update(pbar, 50)
        dvc_destroy(parent_dir)
        progress_update(pbar, 24)

        # venv destroy
        progress_update(pbar, 1)


import tkinter as tk
from tkinter import scrolledtext
from tkinter import PhotoImage

def fdr_documentation():
    command_descriptions = {
        "fdr pull": "Description\n...",
        "fdr push": "Description\n...",
        "fdr show": "Description\n...",
        "fdr doc": "Description\n...",
        "fdr clear": "Description\n...",
        "fdr quit": "Description\n...",
        "fdr log": "Description\n...",
    }

    def show_command_info(command):
        for label in command_labels:
            label.config(bg="SystemButtonFace")  # Reset background color for all labels
        info_text.config(state=tk.NORMAL)  # Make the text widget editable
        info_text.delete(1.0, tk.END)
        info_text.insert(tk.END, f"Command  ---> [ fdr{command[3:]} ]\n\n")
        info_text.insert(tk.END, command_descriptions.get(command, "No documentation available for this command."))
        selected_label = command_labels[commands.index(command)]
        selected_label.config(bg="lightblue")  # Highlight the selected label
        info_text.config(state=tk.DISABLED)  # Disable the text widget to prevent editing

    # Create the main window
    root = tk.Tk()
    root.title("FaradAI Documentation")

    # Create a frame for the command list on the left
    list_frame = tk.Frame(root)
    list_frame.pack(side=tk.LEFT, anchor=tk.N, padx=10, pady=10)

    # List of commands
    commands = ["fdr pull", "fdr push", "fdr show", "fdr doc", "fdr clear", "fdr quit", "fdr log"]

    # Create clickable labels for commands
    command_labels = []
    for command in commands:
        label = tk.Label(list_frame, text=command, cursor="hand2", font=("Times New Roman", 11, "underline"))
        label.pack(anchor=tk.W, pady=(5, 0))
        label.bind("<Button-1>", lambda event, cmd=command: show_command_info(cmd))
        command_labels.append(label)

    # Create a scrolled text widget to display documentation
    info_text = scrolledtext.ScrolledText(root, wrap=tk.WORD, width=80, height=20, font=("Times New Roman", 11))
    info_text.pack(side=tk.TOP, fill=tk.BOTH, expand=True, padx=10, pady=10)
    info_text.config(state=tk.DISABLED)  # Set the text widget to be non-editable
    info_text.config(bg="lightblue", padx=10, pady=10)  # Set yellow background and padding

    # Start GUI window at "fdr pull"
    show_command_info("fdr pull")

    # Start the Tkinter main loop
    root.mainloop()

if __name__ == "__main__":
    fdr_documentation()



import os
import subprocess
import shutil

def create_virtual_environment(target_path):
    """
    Venv creation on target dir & dvc install & progress bar
    """
    parent_dir = os.path.dirname(target_path)
    os.makedirs(parent_dir, exist_ok=True)

    venv_path = os.path.join(parent_dir, 'venv')

    subprocess.run(['python', '-m', 'venv', venv_path], 
                    check=True, 
                    stdout=subprocess.DEVNULL, 
                    stderr=subprocess.DEVNULL)

    
    activate_script = "activate" if os.name == "nt" else "activate"
    activate_path = os.path.join(venv_path, 'Scripts', activate_script)
    subprocess.run([activate_path], 
                    shell=True, 
                    check=True, 
                    stdout=subprocess.DEVNULL, 
                    stderr=subprocess.DEVNULL)
    
    subprocess.run(['pip', 'install', 'dvc'], 
                    check=True, 
                    stdout=subprocess.DEVNULL, 
                    stderr=subprocess.DEVNULL)
    
    return parent_dir

def delete_virtual_environment(parent_dir):
    """
    Venv deletion
    
    """
    venv_path = os.path.join(parent_dir, 'venv')
    if os.path.exists(venv_path):
        shutil.rmtree(venv_path, 
                        onerror=lambda func, 
                        path, 
                        exc_info: None)



# FDR Client Login 

# Libraries
from ..utils.utils import clear
from time import sleep
import sys, os

def auth(un:str, pw:str):
    """
    Login per FDR user 
    """
    if un=="user" and pw=="pass":
        return 200
    else:
        return 404

def auth_denied():
    """
    Message & reload FDR client if auth fails
    """
    print("\nAuthentication failed, restarting FaradAI Client in:")
    for i in range(3):
        print(f'{3-i}')
        sleep(1)
    # clear terminal
    clear()
    # restart script
    os.system(f"{sys.executable} {' '.join(sys.argv)}")


from setuptools import setup, find_packages

# Read dependencies from requirements.txt
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Setup
setup(
    name='fdr-v2',
    version='0.1.0',
    description='Sample fdr lib',
    author='Konstantinos Gyftodimos',
    author_email='konstgyftodimos@gmail.com',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'fdr = client.main:fdr_client',
        ],
    },
    install_requires=requirements
)



# FDR Client Utility Functions

# Libraries
from os import system, name
from ..classes.DatasetManager import DatasetManager
from termcolor import cprint
from tkinter import filedialog
import tkinter as tk
import keyboard
import shutil
import time
import os


def welcome_message():
    """
    FDR welcome message
    """
    cprint('\nWelcome to FaradAI Repository\n', 'cyan', attrs=['bold'])

def clear():
    """
    Clear FDR client
    """
    if name == 'nt':
        # windows
        _ = system('cls')
    else:
        # linux, macos
        _ = system('clear')

def clear_terminal_input():
    """
    Clear terminal input text
    """
    keyboard.press_and_release('esc')

def print_lines_per_terminal_size():
    """
    Get current terminal size
    """
    terminal_width, _ = shutil.get_terminal_size(fallback=(80, 24))
    dash_length = terminal_width - 1
    print(dash_length*"_")

def progress_update(pbar, threshold):
    """
    Smooth progress bar update
    """
    for _ in range(threshold):
        time.sleep(0.05)  
        pbar.update(1)

def dir_picking(message:str):
    """
    GUI to pick dir and return path
    """
    root = tk.Tk()
    root.withdraw() 
    folder_msg = f'{message}'
    folder_path = filedialog.askdirectory(title=folder_msg)
    root.destroy()
    return folder_path

def check_dvc_file(folder_path):
    """
    Check existence of .dvc file in a path
    """
    for filename in os.listdir(folder_path):
        if filename.endswith('.dvc'):
            return True
    return False

def sample_data():
    """
    Sample datasets
    """
    # Sample data
    data = [
        {"ID": "1", "DATASET": "dataset 1", "VERSION": "1.0", "DATE": "5/5"},
        {"ID": "2", "DATASET": "dataset 2", "VERSION": "2.1", "DATE": "8/7"},
        {"ID": "3", "DATASET": "fire_smoke", "VERSION": "8.0", "DATE": "8/7"},
        {"ID": "4", "DATASET": "big_test", "VERSION": "2.0", "DATE": "12/7"}
    ]
    return data

def sample_headers():
    """
    Sample headers
    """
    # Sample headers
    return ["ID", "DATASET", "VERSION", "DATE"]

def get_datasets(header_show:bool, header_color:str):
    """
    FDR get datasets
    """
    dataset_manager = DatasetManager(data=sample_data(), headers=sample_headers())
    if header_color=='cyan' and header_show:
        cprint('\nLoading FaradAI Datasets\n', header_color, attrs=['bold'])
    elif header_show:
        cprint('\nLoading FaradAI Datasets\n', "yellow", attrs=['bold'])

    # progress bar
    dataset_manager.display_loading()

    # datasets
    dataset_manager.display_dataframe()


# FDR Client Command Methods

# Libraries
import os, sys
from ...utils.utils import *
from ...utils.cmd_utils.cmd_fdr import *
from ...utils.cmd_utils.cmd_doc import *
from ...utils.cmd_utils.cmd_logs import CommandLogs
import threading
import keyboard

key_listener_enabled = False

def initial_command_options():
    """
    Show command options if login is successful
    """
    cprint("\nPress key to choose an FDR action: \n", 'cyan', attrs=['blink', 'bold'])
    print("(d) - Datasets")
    print("(l) - Logout")
    print("(q) - Quit")

    # Start the key listener thread
    global key_listener_enabled
    key_listener_enabled = True
    listener_thread = threading.Thread(target=start_listener)
    listener_thread.start()

def dataset_command_options():
    """
    Show command options for dataset handling
    """
    cprint("\nPress key to choose an FDR action: \n", 'cyan', attrs=['blink', 'bold'])
    print("(o) - Operations")
    print("(d) - Datasets")
    print("(l) - Logout")
    print("(q) - Quit")
    enable_key_listener()

def start_listener():
    """
    Start the key listener
    """
    global key_listener_enabled
    r_pressed = False 
    
    while key_listener_enabled:
        if keyboard.is_pressed('d'):
            cmd_datasets()
            r_pressed = True  
        elif keyboard.is_pressed('l'):
            cmd_logout()
        elif keyboard.is_pressed('q'):
            cmd_exit()
        elif r_pressed and keyboard.is_pressed('o'):  
            cmd_operations()

def enable_key_listener():
    """
    Enable the key listener
    """
    global key_listener_enabled
    key_listener_enabled = True

def disable_key_listener():
    """
    Disable the key listener
    """
    global key_listener_enabled
    key_listener_enabled = False

def cmd_exit():
    """
    Exit FDR client
    """
    clear()
    sys.exit()

def cmd_logout():
    """
    Restart FDR client
    """
    clear()
    os.system(f"{sys.executable} {' '.join(sys.argv)}")
    

def cmd_datasets():
    """
    @GET all datasets
    """
    disable_key_listener()
    clear()
    get_datasets(header_show=True, header_color='cyan')
    dataset_command_options()

def cmd_operations():
    """
    FDR Operations on target dataset
    """
    # history log init
    history = CommandLogs()

    # disable key listener
    disable_key_listener()

    # clear ui
    fdr_clear('FDR Dataset Operations')

    # interactive mode
    while True:
        clear_terminal_input()
        user_input = input("\n>> ")
        history.add_text(user_input[0:])    
        parts = user_input.split()

        if parts:
            command = parts[0].lower()

            if command == 'fdr':
                if len(parts) >= 2:
                    subcommand = parts[1].lower()

                    if subcommand == 'pull':
                        pull_args = parts[2:]  
                        fdr_pull(pull_args)
                    elif subcommand == 'push':
                        # handle fdr push
                        fdr_push()
                    elif subcommand == 'menu':
                        cmd_datasets()
                        break
                    elif subcommand == 'clear':
                        fdr_clear('FDR Dataset Operations')
                    elif subcommand == 'show':
                        get_datasets(header_show=False, header_color='white')
                    elif subcommand == 'doc':
                        fdr_documentation()
                    elif user_input.lower() == 'fdr log':
                        cprint(history.get_history(), "yellow", attrs=['bold'])
                    elif user_input.lower() == 'fdr log --save' or user_input.lower() == 'fdr log -s':
                        # handle fdr logs
                        cprint(history.get_history(), "yellow", attrs=['bold'])
                        folder_path = fdr_log_save()
                        if folder_path is not None:
                            file_path = f'{folder_path}/fdr_logs.txt'
                            with open(file_path, 'w') as file:
                                file.write(f'FDR history:\n\n{history.get_history()}')
                    else:
                        print("Invalid argument")
                else:
                    print("Invalid argument")

            else:
                print("Invalid argument")

    # re-enable key listener
    enable_key_listener()

if __name__ == "__main__":
    initial_command_options()



# Libraries
from time import sleep
import shutil
from ..classes.CustomTqdm import CustomTqdm
from colorama import Fore
import pandas as pd

# Dataset Manager Class
class DatasetManager:
    def __init__(self, data, headers):
        self.df = pd.DataFrame(data, columns=headers)
        self.terminal_width, _ = shutil.get_terminal_size(fallback=(80, 24))
        self.dash_length = self.terminal_width - 1

    def display_loading(self):
        with CustomTqdm(total=7, 
                        ncols=self.terminal_width, 
                        bar_format="{l_bar}" + Fore.WHITE + "{bar}" + Fore.RESET + "{r_bar}") as pbar:
            for _ in range(7):
                sleep(0.2)
                pbar.update(1)

    def get_dataset_names(self):
        return self.df["DATASET"].tolist()

    def display_dataframe(self):
        pd.set_option("display.max_columns", None)
        print(self.dash_length * "_")
        print(self.df.to_string(index=False))
        print(self.dash_length * "_")

if __name__ == "__main__":
    # Sample data
    data = [
        {"ID": "1", "DATASET": "dataset 1", "VERSION": "1.0", "DATE": "5/5"},
        {"ID": "2", "DATASET": "dataset 2", "VERSION": "2.1", "DATE": "8/7"},
        {"ID": "3", "DATASET": "fire_smoke", "VERSION": "8.0", "DATE": "8/7"}
    ]

    # Corresponding headers
    headers = ["ID", "DATASET", "VERSION", "DATE"]

    manager = DatasetManager(data, headers)
    manager.display_loading()
    manager.display_dataframe()
